from pathlib import Path


def safe_path(base, path):
    base = Path(base).resolve()
    absolute = Path(base).joinpath(path).resolve()
    absolute.relative_to(Path(base).resolve())
    return absolute



#!/usr/bin/env python

"""siembra

Usage:
  siembra new <package> <output-directory>

Options:
  -h --help     Show this screen.
"""

from pathlib import Path

import yaml
import docker  # type: ignore
from docopt import docopt  # type: ignore

from executors.docker import DockerConfig, DockerExecutor
from executors.file import (
    DirectoriesConfig,
    FileExecutorConfig,
    CopyExecutor,
    MoveExecutor,
    DeleteExecutor,
)
from executors.template import TemplateExecutor
from parameters import parse_parameters_dictionary

class Siembra:
    def __init__(self, spec):
        pass 

    def run(self, package):



def new(package:  Path, output_directory: Path) -> None:
    spec = yaml.load(open(package / 'siembra.yml'), Loader=yaml.Loader)

    parameters = parse_parameters_dictionary(spec.get('parameters', {}))

    directories_config = DirectoriesConfig(
        package=package,
        output=output_directory,
    )

    for step in spec['steps']:
        if step.get('docker'):
            config = DockerConfig.from_yml(step.get('docker'))

            DockerExecutor(config=config).run(volume=directories_config.output)
        elif step.get('copy'):
            CopyExecutor(config=directories_config).run(
                runtime=FileExecutorConfig.from_yml(step['copy'])
            )
        elif step.get('move'):
            MoveExecutor(config=directories_config).run(
                runtime=FileExecutorConfig.from_yml(step['move'])
            )
        elif step.get('delete'):
            DeleteExecutor(config=directories_config).run(step.get('delete'))
        elif step.get('template'):
            TemplateExecutor(config=parameters).run(
                directories_config.output / step.get('template', {}).get('file'))


def main() -> None:
    args = docopt(__doc__, version='0.1')

    if args.get('new'):
        new(package=Path(args['<package>']),
            output_directory=Path(args['<output-directory>']))


if __name__ == '__main__':
    main()



from uuid import uuid4
from pathlib import Path


class WorkingFile:
    def __init__(self, package: Path, name: str = 'siembra-temporary'):
        # FIXME: Check that path is a direct descendant of `package`
        path = (package / name).resolve()
        if path.exists() and not path.is_dir():
            path.unlink()
        path.mkdir(parents=True, exist_ok=True)

        self.working_directory = path

    def create(self):
        max_retries = 3
        for try_count in range(max_retries):
            try:
                path = self.working_directory / str(uuid4())
                return open(path, 'x')
            except FileExistsError as e:
                if try_count != max_retries - 1:
                    continue
                raise e



from string import Template
from pathlib import Path


class TemplateExecutor:
    def __init__(self, config: dict):
        self.config = config

    def run(self, path: Path) -> None:
        template = Template(open(path).read())
        open(path, 'w').write(
            template.substitute(self.config)
        )



import os
import shutil
from typing import Tuple
from collections.abc import Callable
from pathlib import Path

from .safe_path import safe_path


class InvalidPackagePath(Exception):
    def __init__(self, package: Path, path: Path):
        self.package = package
        self.path = path

        super().__init__(f'Path "{path}" is not a child of "{package}"')


class InvalidOutputPath(Exception):
    def __init__(self, output: Path, path: Path):
        self.output = output
        self.path = path

        super().__init__(f'Path "{path}" is not a child of "{output}"')


class DirectoriesConfig:
    def __init__(self, package: Path, output: Path):
        self.package = package.resolve()
        self.output = output.resolve()


class FileExecutorConfig:
    def __init__(self,
                 dest: Path,
                 from_package: Path = None,
                 from_output: Path = None):
        if from_package is not None and from_output is not None:
            # FIXME: Raise a proper exception for this
            raise Exception('ERROR')

        self.from_package = from_package if from_package else None
        self.from_output = from_output if from_output else None
        self.dest = dest

    @classmethod
    def from_yml(cls, yml):
        from_pkg = yml.get('from-pkg')
        from_output = yml.get('from')

        return cls(
            dest=Path(yml['to']),
            from_package=Path(from_pkg) if from_pkg else None,
            from_output=Path(from_output) if from_output else None,
        )


class FileExecutor:
    def __init__(self, config: DirectoriesConfig):
        self.config = config

    def mkdir(self, path):
        if not path.is_dir():
            path = path.parent
        path.mkdir(parents=True, exist_ok=True)

    def get_paths(self, runtime: FileExecutorConfig) -> Tuple[Path, Path]:
        if runtime.from_package:
            try:
                src = safe_path(self.config.package,
                                runtime.from_package)
            except ValueError:
                raise InvalidPackagePath(package=self.config.package,
                                         path=runtime.from_package)

        if runtime.from_output:
            try:
                src = safe_path(self.config.output,
                                runtime.from_output)
            except ValueError:
                raise InvalidOutputPath(output=self.config.output,
                                        path=runtime.from_output)

        try:
            dst = safe_path(self.config.output,
                            runtime.dest)
        except ValueError:
            raise InvalidOutputPath(output=self.config.output,
                                    path=runtime.dest)

        return (src, dst)

    def run(self,
            runtime: FileExecutorConfig,
            apply_fn) -> None:
        # import pdb
        # pdb.set_trace()

        src, dst = self.get_paths(runtime=runtime)
        self.mkdir(dst)
        apply_fn(src, dst)


class CopyExecutor:
    def __init__(self, config: DirectoriesConfig):
        self.config = config

    def run(self, runtime: FileExecutorConfig) -> None:
        def copy(src, dst):
            if src.is_dir():
                shutil.copytree(src, dst)
            else:
                shutil.copy(src, dst)

        FileExecutor(config=self.config).run(runtime=runtime,
                                             apply_fn=copy)


class MoveExecutor:
    def __init__(self, config: DirectoriesConfig):
        self.config = config

    def run(self, runtime: FileExecutorConfig) -> None:
        FileExecutor(config=self.config).run(
            runtime=runtime,
            apply_fn=lambda src, dst: src.replace(dst)
        )


class DeleteExecutor:
    def __init__(self, config: DirectoriesConfig):
        self.config = config

    def run(self, path: Path) -> None:
        path = safe_path(self.config.output, path)
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink(missing_ok=True)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="siembra",
    version="0.1.0",
    author="Juan Cabrera",
    author_email="juan@jjcp.dev",
    description="Description",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mulitadev/siembra",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'siembra=src.siembra:main',
        ],
    },
)


import sys
import subprocess
from typing import List
from pathlib import Path

from .working_file import WorkingFile


class DockerConfig:
    def __init__(self, image: str, commands: List[str]):
        self.image = image
        self.commands = commands

    @classmethod
    def from_yml(cls, yml: dict):
        return cls(
            image=yml['image'],
            commands=yml['commands'],
        )


class DockerExecutor:
    def __init__(self, config: DockerConfig):
        self.config = config

    def run(self, volume: Path) -> None:
        # FIXME: This should be implemented using Docker's
        #        python sdk

        wf = WorkingFile(package=Path('.'), name='siembra-temporary')

        for command in self.config.commands:
            sh_file = wf.create()
            sh_path = Path(sh_file.name).resolve()

            sh_file.write(f'{command}')
            sh_file.close()

            subprocess.run([
                'docker', 'run', '--rm', '-ti',
                '-v', f'{volume.resolve()}:/app',
                '-v', f'{sh_path}:/tmp/run.sh',
                '-w', '/app',
                '--entrypoint',
                'bash',
                f'{self.config.image}',
                '/tmp/run.sh'
            ])
